/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.layout;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_INSETS;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_HEIGHT;
import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_SHAPE_MINIMUM_WIDTH;
import static org.fortiss.tooling.base.layout.LayoutKeyConstants.CONNECTOR_OFFSET_ORIENTATION;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getAbsolute;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.getNodeBounds;
import static org.fortiss.tooling.base.ui.utils.LayoutDataUtils.setConnectorOffsetOrientation;
import static org.fortiss.tooling.base.ui.utils.RectangleLayoutUtils.calculateRectangleBorderOrientation;
import static org.fortiss.tooling.base.ui.utils.RectangleLayoutUtils.clampRectangularShapeSize;
import static org.fortiss.tooling.base.ui.utils.SnapToGridUtils.snapToGrid;

import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Viewport;
import org.eclipse.draw2d.geometry.Dimension;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.editparts.ScalableRootEditPart;
import org.eclipse.gef.requests.ChangeBoundsRequest;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.OffsetOrientation;
import org.fortiss.tooling.base.ui.editpart.ElementEditPartBase;
import org.fortiss.tooling.base.ui.editpart.StickyConnectorEditPartBase;
import org.fortiss.tooling.base.ui.utils.LayoutDataUtils;

/**
 * This layout configuration implements the rectangles with grid flavor. The
 * diagram has a grid twice the resolution of a connector shape size (e.g. a
 * connector spans a 2x2 grid area). Its elements and free connectors snap to
 * the grid. Elements use a rectangular shape and sticky connectors are centered
 * on the border of their parent element. Connections use straight lines with
 * bend points also snapping to the grid.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 412BCCEA0E2EF9D3B58DC15887A4D43D
 */
public class RectangleGridLayoutDiagramConfiguration extends
		DiagramLayoutConfigurationBase {

	/** Shared instance. */
	public static final RectangleGridLayoutDiagramConfiguration INSTANCE = new RectangleGridLayoutDiagramConfiguration();

	/** The minimal size. */
	private static final Dimension MINIMUM_SIZE = new Dimension(
			DEFAULT_SHAPE_MINIMUM_WIDTH, DEFAULT_SHAPE_MINIMUM_HEIGHT);

	/** {@inheritDoc} */
	@Override
	public boolean useGrid() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public void performBendPointLayout(Point p) {
		p.x = snapToGrid(p.x);
		p.y = snapToGrid(p.y);
	}

	/** {@inheritDoc} */
	@Override
	public Dimension getMinimalElementSize() {
		return MINIMUM_SIZE;
	}

	/** {@inheritDoc} */
	@Override
	public void performStickyConnectorMove(
			StickyConnectorEditPartBase<? extends ILayoutedModelElement> connectorEditPart,
			ChangeBoundsRequest request) {
		// get extended data altered by feedback below
		OffsetOrientation orient = (OffsetOrientation) request
				.getExtendedData().get(CONNECTOR_OFFSET_ORIENTATION);
		setConnectorOffsetOrientation(connectorEditPart.getModel(), orient);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public void alterStickyConnectorMoveFeedbackRequest(
			ChangeBoundsRequest request,
			StickyConnectorEditPartBase<? extends ILayoutedModelElement> connectorEditPart,
			IFigure connectorFigure) {
		// get the bounds of the parent
		Rectangle containerBounds = getNodeBounds(connectorEditPart
				.getParentEditPart().getModel());
		// get the child position relative to the parent
		Point childPos = connectorFigure.getBounds().getLocation()
				.getTranslated(containerBounds.getLocation().getNegated());
		// get the current mouse position
		Point mousePos = request.getLocation();
		// fix offset in mouse position because of possibly scrolled editing
		// window
		Point clientAreaOffset = new Point(
				((Viewport) ((ScalableRootEditPart) connectorEditPart
						.getParentEditPart().getRoot()).getFigure())
						.getClientArea().getLocation());
		mousePos.translate(clientAreaOffset);
		// get the mouse position relative to the parent
		Point relativePos = mousePos.getTranslated(containerBounds
				.getLocation().getNegated());

		// do border clamping
		OffsetOrientation orient = calculateRectangleBorderOrientation(
				relativePos, containerBounds.getSize(), DEFAULT_SHAPE_INSETS);
		// snap to grid
		orient.setOffset(snapToGrid(orient.getOffset()));
		// use the calculated value for positioning
		org.fortiss.tooling.base.model.layout.Dimension d = LayoutDataUtils
				.getNodeSize(connectorEditPart.getParentEditPart().getModel());
		org.fortiss.tooling.base.model.layout.Point p = getAbsolute(orient, d,
				DEFAULT_SHAPE_INSETS);
		relativePos.setLocation(p.getX(), p.getY());

		// store the orientation in the request so that the command can use
		// it
		request.getExtendedData().put(CONNECTOR_OFFSET_ORIENTATION, orient);
		// MoveDelta needs positions relative to the child
		relativePos.translate(childPos.getNegated());

		request.setMoveDelta(relativePos);
	}

	/** {@inheritDoc} */
	@Override
	public void alterElementMoveAndResizeFeedbackRequest(
			ChangeBoundsRequest request,
			ElementEditPartBase<? extends ILayoutedModelElement> elementEditPart) {
		Dimension oldSizeDelta = request.getSizeDelta();
		Rectangle bounds = elementEditPart.getFigure().getBounds();
		Dimension oldSize = new Dimension(bounds.width + oldSizeDelta.width,
				bounds.height + oldSizeDelta.height);
		Dimension minimal = getMinimalElementSize();
		Dimension newSize = clampRectangularShapeSize(oldSize, minimal.width,
				minimal.height);
		Dimension newSizeDelta = new Dimension(newSize.width - bounds.width,
				newSize.height - bounds.height);
		request.setSizeDelta(newSizeDelta);
	}
}
